#include "common.h"

void libmemcached_free(memcached_st *ptr, void *mem)
{
  (void) ptr;
  free(mem);
}

void *libmemcached_malloc(memcached_st *ptr, size_t size)
{
  (void) ptr;
  return malloc(size);
}

void *libmemcached_realloc(memcached_st *ptr, void *mem, size_t size)
{
  (void) ptr;
  return realloc(mem, size);
}

void *libmemcached_calloc(memcached_st *ptr, size_t nelem, size_t size)
{
  if (ptr->call_malloc != libmemcached_malloc)
  {
     void *ret = libmemcached_malloc(ptr, nelem * size);
     if (ret != NULL) 
       memset(ret, 0, nelem * size);

     return ret;
  }

  return calloc(nelem, size);
}

memcached_return memcached_set_memory_allocators(memcached_st *ptr,
                                                 memcached_malloc_function mem_malloc,
                                                 memcached_free_function mem_free,
                                                 memcached_realloc_function mem_realloc,
                                                 memcached_calloc_function mem_calloc)
{
  /* All should be set, or none should be set */
  if (mem_malloc == NULL && mem_free == NULL && mem_realloc == NULL && mem_calloc == NULL) 
  {
    ptr->call_malloc= libmemcached_malloc;
    ptr->call_free= libmemcached_free;
    ptr->call_realloc= libmemcached_realloc;
    ptr->call_calloc= libmemcached_calloc;
  }
  else if (mem_malloc == NULL || mem_free == NULL || mem_realloc == NULL || mem_calloc == NULL)
    return MEMCACHED_FAILURE;
  else
  {
    ptr->call_malloc= mem_malloc;
    ptr->call_free= mem_free;
    ptr->call_realloc= mem_realloc;
    ptr->call_calloc= mem_calloc;
  }

  return MEMCACHED_SUCCESS;
}

void memcached_get_memory_allocators(memcached_st *ptr,
                                     memcached_malloc_function *mem_malloc,
                                     memcached_free_function *mem_free,
                                     memcached_realloc_function *mem_realloc,
                                     memcached_calloc_function *mem_calloc)
{
   *mem_malloc= ptr->call_malloc;
   *mem_free= ptr->call_free;
   *mem_realloc= ptr->call_realloc;
   *mem_calloc= ptr->call_calloc;
}

/***************************
 * shm allocators
 ***************************/

/* Some temporary defines */
#define SHMKEY "/memc-test"
#define SHMMODE 0600
#define SHMMAGIC 0xBEEF
/* end defines */

/* TODO: real checks for success and whatnot. Just filling in stuff here still
 * at this stage of the game. */
int libmemcached_shm_init(memcached_st *ptr, size_t amount){
	/* Allocate the shared memory area. Start off using POSIX shared mem
 	 * for relative ease of use, but we should also do SysV shared mem
	 * down the road. */
	struct stat sb;
	amount += sizeof(struct memcached_shm);
	ptr->sfd = shm_open(SHMKEY, O_RDWR | O_CREAT, SHMMODE);
	if(ptr->sfd == -1){
		perror("shm_open");
		}
	if(fstat(ptr->sfd, &sb) == -1){
		perror("fstat");
		exit(-1);
		}
	if((unsigned) sb.st_size < amount){
		/* check this later */
		if(ftruncate(ptr->sfd, amount) == -1){
			perror("ftruncate");
			fprintf(stderr, "Something went wrong with ftruncate!\n");
			exit(-1);
			}
		}
	/* Mmap control structure for allocated shared memory */
	/* Remains to be seen if this is the best way of going about it, but
	 * I'm not coming up with anything better yet. Of course, this is kind
	 * of a new area for me. */
	/* sizeof(memcached_shm) */
	ptr->shm_area = mmap((caddr_t)0, amount, PROT_READ | PROT_WRITE, MAP_SHARED, ptr->sfd, 0);
	if (ptr->shm_area == MAP_FAILED){
		perror("mmap");
		exit(-1);
		}
	/* allocated the memcached_shm struct. Increase reference */
	ptr->shm_area->referencing++;
	/* set the size allocated */
	ptr->shm_area->shm_size = amount;
	/* Set the rest of the memory to ptr->shm_area->shm */
	/* and... jump down the line. Maybe it will even work. */
	ptr->shm_area->shm = (ptr->shm_area + sizeof(struct memcached_shm)); 
		
	return(0);
	}

int libmemcached_shm_close(memcached_st *ptr){
	/* Structures for keeping track of what we've allocated. This may yet
	 * not be what we want to do. */
	int rchk = --(ptr->shm_area->referencing);
	size_t msize = ptr->shm_area->shm_size;
	if (munmap(ptr->shm_area, msize) == -1)
		perror("munmap");
	if (rchk == 0)
		shm_unlink(SHMKEY); /* unlink *if* we're the last using it. */
	close(ptr->sfd);
	return(0);
	}

