/* -*- Mode: C; tab-width: 2; c-basic-offset: 2; indent-tabs-mode: nil -*- */
#include "common.h"

memcached_return memcached_delete(memcached_st *ptr, const char *key, size_t key_length,
                                  time_t expiration)
{
  return memcached_delete_by_key(ptr, key, key_length,
                                 key, key_length, expiration);
}

memcached_return memcached_delete_by_key(memcached_st *ptr,
                                         const char *master_key,
                                         size_t master_key_length,
                                         const char *key,
                                         size_t key_length,
                                         time_t expiration)
{
  unlikely (expiration)
    return MEMCACHED_INVALID_ARGUMENTS;

  memcached_return rc;
  unsigned int server_key;

  LIBMEMCACHED_MEMCACHED_DELETE_START();

  unlikely (ptr->hosts == NULL || ptr->number_of_hosts == 0)
    return MEMCACHED_NO_SERVERS;

  server_key= memcached_generate_hash(ptr, master_key, master_key_length);
  rc= ptr->protocol_hooks->delete(ptr, server_key, key, key_length);
  if (rc == MEMCACHED_SUCCESS && ptr->delete_trigger)
    ptr->delete_trigger(ptr, key, key_length);

  LIBMEMCACHED_MEMCACHED_DELETE_END();
  return rc;
}

static inline memcached_return binary_delete(memcached_st *ptr,
                                             unsigned int server_key,
                                             const char *key,
                                             size_t key_length,
                                             uint8_t flush)
{
  protocol_binary_request_delete request= {
     .message.header.request= {
        .magic= PROTOCOL_BINARY_REQ,
        .opcode= (ptr->flags & MEM_NOREPLY) ? PROTOCOL_BINARY_CMD_DELETEQ :
                                              PROTOCOL_BINARY_CMD_DELETE,
        .keylen= htons((uint16_t)key_length),
        .datatype= PROTOCOL_BINARY_RAW_BYTES,
        .bodylen= htonl((uint32_t) key_length)
     }
  };

  if (ptr->flags & MEM_USE_UDP && !flush)
  {
    size_t cmd_size= sizeof(request.bytes) + key_length;
    if (cmd_size > MAX_UDP_DATAGRAM_LENGTH - UDP_DATAGRAM_HEADER_LENGTH)
      return MEMCACHED_WRITE_FAILURE;
    if (cmd_size + ptr->hosts[server_key].write_buffer_offset > MAX_UDP_DATAGRAM_LENGTH)
      memcached_io_write(&ptr->hosts[server_key], NULL, 0, 1);
  }

  memcached_return rc= MEMCACHED_SUCCESS;

  if ((memcached_do(&ptr->hosts[server_key], request.bytes,
                    sizeof(request.bytes), 0) != MEMCACHED_SUCCESS) ||
      (memcached_io_write(&ptr->hosts[server_key], key,
                          key_length, (char) flush) == -1))
  {
    memcached_io_reset(&ptr->hosts[server_key]);
    rc= MEMCACHED_WRITE_FAILURE;
  }

  unlikely (ptr->number_of_replicas > 0)
  {
    request.message.header.request.opcode= PROTOCOL_BINARY_CMD_DELETEQ;

    for (uint32_t x= 0; x < ptr->number_of_replicas; ++x)
    {
      ++server_key;
      if (server_key == ptr->number_of_hosts)
        server_key= 0;

      memcached_server_st* server= &ptr->hosts[server_key];
      if ((memcached_do(server, (const char*)request.bytes,
                        sizeof(request.bytes), 0) != MEMCACHED_SUCCESS) ||
          (memcached_io_write(server, key, key_length, (char) flush) == -1))
        memcached_io_reset(server);
      else
        memcached_server_response_decrement(server);
    }
  }

  return rc;
}

memcached_return memcached_binary_delete(memcached_st *ptr,
                                         unsigned int server_key,
                                         const char *key,
                                         size_t key_length)
{
  memcached_return rc= memcached_validate_key_length(key_length, true);
  unlikely (rc != MEMCACHED_SUCCESS)
    return rc;
  uint8_t to_write= (uint8_t)((ptr->flags & MEM_BUFFER_REQUESTS) ? 0 : 1);
  bool no_reply= (ptr->flags & MEM_NOREPLY);

  rc= binary_delete(ptr, server_key, key, key_length, to_write);
  if (rc == MEMCACHED_SUCCESS)
  {
    if ((ptr->flags & MEM_BUFFER_REQUESTS))
      rc= MEMCACHED_BUFFERED;
    else if (!no_reply)
    {
      char buffer[MEMCACHED_DEFAULT_COMMAND_SIZE];
      rc= memcached_response(&ptr->hosts[server_key], buffer,
                             sizeof(buffer), NULL);
      if (rc == MEMCACHED_DELETED)
        rc= MEMCACHED_SUCCESS;
    }
  }

  return rc;
}

memcached_return memcached_internal_delete(memcached_st *ptr,
                                           unsigned int server_key,
                                           const char *key,
                                           size_t key_length)
{
  (void)server_key;
  memcached_return rc= memcached_validate_key_length(key_length, false);
  unlikely (rc != MEMCACHED_SUCCESS)
    return rc;

  return internal_delete(ptr, key, key_length);
}


memcached_return memcached_ascii_delete(memcached_st *ptr,
                                        unsigned int server_key,
                                        const char *key,
                                        size_t key_length)
{
  memcached_return rc= memcached_validate_key_length(key_length, false);
  unlikely (rc != MEMCACHED_SUCCESS)
    return rc;

  uint8_t to_write= (uint8_t)((ptr->flags & MEM_BUFFER_REQUESTS) ? 0 : 1);
  bool no_reply= (ptr->flags & MEM_NOREPLY);
  char buffer[MEMCACHED_DEFAULT_COMMAND_SIZE];
  size_t send_length= (size_t) snprintf(buffer, sizeof(buffer),
                                        "delete %s%.*s%s\r\n",
                                        ptr->prefix_key,
                                        (int)key_length, key,
                                        no_reply ? " noreply" :"");

  if (send_length >= MEMCACHED_DEFAULT_COMMAND_SIZE)
  {
    rc= MEMCACHED_WRITE_FAILURE;
    goto error;
  }

  if (ptr->flags & MEM_USE_UDP && !to_write)
  {
    if (send_length > MAX_UDP_DATAGRAM_LENGTH - UDP_DATAGRAM_HEADER_LENGTH)
    {
      return MEMCACHED_WRITE_FAILURE;
    }

    if (send_length + ptr->hosts[server_key].write_buffer_offset > MAX_UDP_DATAGRAM_LENGTH)
    {
      memcached_io_write(&ptr->hosts[server_key], NULL, 0, 1);
    }
  }

  rc= memcached_do(&ptr->hosts[server_key], buffer, send_length, to_write);

  if (rc != MEMCACHED_SUCCESS)
    goto error;

  if (!to_write)
    rc= MEMCACHED_BUFFERED;
  else if (!no_reply)
  {
    rc= memcached_response(&ptr->hosts[server_key], buffer, MEMCACHED_DEFAULT_COMMAND_SIZE, NULL);
    if (rc == MEMCACHED_DELETED)
      rc= MEMCACHED_SUCCESS;
  }

  if (rc == MEMCACHED_SUCCESS && ptr->delete_trigger)
    ptr->delete_trigger(ptr, key, key_length);

error:
  LIBMEMCACHED_MEMCACHED_DELETE_END();
  return rc;
}
