#ifndef LIBMEMCACHED_PROTOCOL_HOOKS_H
#define LIBMEMCACHED_PROTOCOL_HOOKS_H

typedef enum {
   LIBMEMCACHED_ASCII_PROTOCOL,
   LIBMEMCACHED_BINARY_PROTOCOL,
   LIBMEMCACHED_EMBEDDED_PROTOCOL
} libmemcached_protocol;

typedef memcached_return (*clone_hook)(memcached_st *org, memcached_st *clone);
typedef void (*free_hook)(memcached_st *instance);


typedef memcached_return (*send_storage)(memcached_st *ptr,
                                         const char *master_key,
                                         size_t master_key_length,
                                         const char *key,
                                         size_t key_length,
                                         const char *value,
                                         size_t value_length,
                                         time_t expiration,
                                         uint32_t flags,
                                         uint64_t cas,
                                         memcached_storage_action verb);

typedef memcached_return (*send_flush)(memcached_st *ptr, time_t exp);
typedef void (*send_quit_server)(memcached_server_st *ptr, uint8_t io_death);

typedef memcached_return (*do_arithmetic)(memcached_st *ptr, bool incr,
                                          const char *mkey, size_t mkey_length,
                                          const char *key, size_t key_length,
                                          uint32_t offset, uint64_t *value);
typedef memcached_return (*do_arithmetic_with_initial)(memcached_st *ptr,
                                                       bool incr,
                                                       const char *mkey,
                                                       size_t mkey_length,
                                                       const char *key,
                                                       size_t key_length,
                                                       uint64_t offset,
                                                       uint64_t initial,
                                                       time_t expiration,
                                                       uint64_t *value);

typedef memcached_return (*send_delete)(memcached_st *ptr,
                                        unsigned int server_key,
                                        const char *key,
                                        size_t key_length);

typedef memcached_return (*stats_fetch)(memcached_st *ptr,
                                        memcached_stat_st *memc_stat,
                                        const char *args,
                                        unsigned int server_key);

typedef memcached_return (*version)(memcached_st *ptr);

typedef memcached_return (*send_mget)(memcached_st *ptr,
                                      const char *master_key,
                                      size_t master_key_length,
                                      const char * const *keys,
                                      const size_t *key_length,
                                      size_t number_of_keys,
                                      bool mget_mode);

typedef memcached_return (*dump_cache)(memcached_st *ptr,
                                       memcached_dump_func *callback,
                                       void *context,
                                       uint32_t number_of_callbacks);


struct memcached_protocol_hooks {
   send_storage storage;
   send_flush flush;
   send_quit_server quit_server;
   do_arithmetic arithmetic;
   do_arithmetic_with_initial arithmetic_with_initial;
   send_delete delete;
   stats_fetch fetch_stats;
   version version;
   send_mget mget;
   dump_cache dump;
   clone_hook clone_hook;
   free_hook free_hook;
};


LIBMEMCACHED_LOCAL
struct memcached_protocol_hooks* get_protocol_hooks(libmemcached_protocol p);

LIBMEMCACHED_LOCAL
memcached_return memcached_binary_storage(memcached_st *ptr,
                                          const char *master_key,
                                          size_t master_key_length,
                                          const char *key,
                                          size_t key_length,
                                          const char *value,
                                          size_t value_length,
                                          time_t expiration,
                                          uint32_t flags,
                                          uint64_t cas,
                                          memcached_storage_action verb);

LIBMEMCACHED_LOCAL
memcached_return memcached_ascii_storage(memcached_st *ptr,
                                         const char *master_key,
                                         size_t master_key_length,
                                         const char *key,
                                         size_t key_length,
                                         const char *value,
                                         size_t value_length,
                                         time_t expiration,
                                         uint32_t flags,
                                         uint64_t cas,
                                         memcached_storage_action verb);


LIBMEMCACHED_LOCAL
memcached_return memcached_internal_storage(memcached_st *ptr,
                                            const char *master_key,
                                            size_t master_key_length,
                                            const char *key,
                                            size_t key_length,
                                            const char *value,
                                            size_t value_length,
                                            time_t expiration,
                                            uint32_t flags,
                                            uint64_t cas,
                                            memcached_storage_action verb);

LIBMEMCACHED_LOCAL
memcached_return memcached_ascii_flush(memcached_st *ptr, time_t expiration);

LIBMEMCACHED_LOCAL
memcached_return memcached_binary_flush(memcached_st *ptr, time_t expiration);

LIBMEMCACHED_LOCAL
memcached_return memcached_internal_flush(memcached_st *ptr, time_t expiration);

LIBMEMCACHED_LOCAL
void memcached_ascii_quit_server(memcached_server_st *ptr, uint8_t io_death);

LIBMEMCACHED_LOCAL
void memcached_binary_quit_server(memcached_server_st *ptr, uint8_t io_death);

LIBMEMCACHED_LOCAL
void memcached_internal_quit_server(memcached_server_st *ptr, uint8_t io_death);

LIBMEMCACHED_LOCAL
memcached_return memcached_ascii_arithmetic(memcached_st *ptr, bool incr,
                                            const char *mkey,
                                            size_t mkey_length,
                                            const char *key, size_t key_length,
                                            uint32_t offset,
                                            uint64_t *value);

LIBMEMCACHED_LOCAL
memcached_return memcached_binary_arithmetic(memcached_st *ptr, bool incr,
                                             const char *mkey,
                                             size_t mkey_length,
                                             const char *key, size_t key_length,
                                             uint32_t offset,
                                             uint64_t *value);

LIBMEMCACHED_LOCAL
memcached_return memcached_internal_arithmetic(memcached_st *ptr, bool incr,
                                               const char *mkey,
                                               size_t mkey_length,
                                               const char *key,
                                               size_t key_length,
                                               uint32_t offset,
                                               uint64_t *value);

LIBMEMCACHED_LOCAL
memcached_return memcached_ascii_arithmetic_with_initial(memcached_st *ptr,
                                                         bool incr,
                                                         const char *mkey,
                                                         size_t mkey_length,
                                                         const char *key,
                                                         size_t key_length,
                                                         uint64_t offset,
                                                         uint64_t initial,
                                                         time_t expiration,
                                                         uint64_t *value);
LIBMEMCACHED_LOCAL
memcached_return memcached_binary_arithmetic_with_initial(memcached_st *ptr,
                                                          bool incr,
                                                          const char *mkey,
                                                          size_t mkey_length,
                                                          const char *key,
                                                          size_t key_length,
                                                          uint64_t offset,
                                                          uint64_t initial,
                                                          time_t expiration,
                                                          uint64_t *value);
LIBMEMCACHED_LOCAL
memcached_return memcached_internal_arithmetic_with_initial(memcached_st *ptr,
                                                            bool incr,
                                                            const char *mkey,
                                                            size_t mkey_length,
                                                            const char *key,
                                                            size_t key_length,
                                                            uint64_t offset,
                                                            uint64_t initial,
                                                            time_t expiration,
                                                            uint64_t *value);


LIBMEMCACHED_LOCAL
memcached_return memcached_ascii_delete(memcached_st *ptr,
                                        unsigned int server_key,
                                        const char *key,
                                        size_t key_length);

LIBMEMCACHED_LOCAL
memcached_return memcached_binary_delete(memcached_st *ptr,
                                         unsigned int server_key,
                                         const char *key,
                                         size_t key_length);

LIBMEMCACHED_LOCAL
memcached_return memcached_internal_delete(memcached_st *ptr,
                                           unsigned int server_key,
                                           const char *key,
                                           size_t key_length);


LIBMEMCACHED_LOCAL
memcached_return memcached_ascii_stats_fetch(memcached_st *ptr,
                                             memcached_stat_st *memc_stat,
                                             const char *args,
                                             unsigned int server_key);

LIBMEMCACHED_LOCAL
memcached_return memcached_binary_stats_fetch(memcached_st *ptr,
                                              memcached_stat_st *memc_stat,
                                              const char *args,
                                              unsigned int server_key);


LIBMEMCACHED_LOCAL
memcached_return memcached_ascii_version(memcached_st *ptr);

LIBMEMCACHED_LOCAL
memcached_return memcached_binary_version(memcached_st *ptr);

LIBMEMCACHED_LOCAL
memcached_return memcached_internal_version(memcached_st *ptr);


LIBMEMCACHED_LOCAL
memcached_return memcached_ascii_mget(memcached_st *ptr,
                                      const char *master_key,
                                      size_t master_key_length,
                                      const char * const *keys,
                                      const size_t *key_length,
                                      size_t number_of_keys,
                                      bool mget_mode);

LIBMEMCACHED_LOCAL
memcached_return memcached_binary_mget(memcached_st *ptr,
                                       const char *master_key,
                                       size_t master_key_length,
                                       const char * const *keys,
                                       const size_t *key_length,
                                       size_t number_of_keys,
                                       bool mget_mode);

LIBMEMCACHED_LOCAL
memcached_return memcached_internal_mget(memcached_st *ptr,
                                         const char *master_key,
                                         size_t master_key_length,
                                         const char * const *keys,
                                         const size_t *key_length,
                                         size_t number_of_keys,
                                         bool mget_mode);

LIBMEMCACHED_LOCAL
memcached_return memcached_ascii_dump(memcached_st *ptr,
                                      memcached_dump_func *callback,
                                      void *context,
                                      uint32_t number_of_callbacks);

LIBMEMCACHED_LOCAL
memcached_return memcached_binary_dump(memcached_st *ptr,
                                       memcached_dump_func *callback,
                                       void *context,
                                       uint32_t number_of_callbacks);

LIBMEMCACHED_LOCAL
memcached_return memcached_internal_dump(memcached_st *ptr,
                                         memcached_dump_func *callback,
                                         void *context,
                                         uint32_t number_of_callbacks);

LIBMEMCACHED_LOCAL
memcached_return memcached_internal_clone_hook(memcached_st *org,
                                               memcached_st *clone);

LIBMEMCACHED_LOCAL
void memcached_internal_free_hook(memcached_st *instance);

LIBMEMCACHED_LOCAL
memcached_return memcached_noop_clone_hook(memcached_st *org,
                                           memcached_st *clone);

LIBMEMCACHED_LOCAL
void memcached_noop_free_hook(memcached_st *instance);

#endif /* LIBMEMCACHED_PROTOCOL_HOOKS_H */
